from . import tfw_util as _tfw_util

HAS_TORCH = _tfw_util.HAS_TORCH
HAS_YOLOV5_FACE = _tfw_util.HAS_YOLOV5_FACE
HAS_REQUESTS = _tfw_util.HAS_REQUESTS
HAS_TQDM = _tfw_util.HAS_TQDM

__all__ = [
    "HAS_TORCH",
    "HAS_YOLOV5_FACE",
    "HAS_REQUESTS",
    "HAS_TQDM",
    *_tfw_util.__all__,
]

for _name in _tfw_util.__all__:
    globals()[_name] = getattr(_tfw_util, _name)

del _tfw_util, _name
